#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe both DMs and guild channels on Discord."""

from __future__ import annotations

__all__ = [
    "ChannelType",
    "PermissionOverwrite",
    "PermissionOverwriteType",
    "PartialChannel",
    "DMChannel",
    "GroupDMChannel",
    "GuildCategory",
    "GuildChannel",
    "GuildTextChannel",
    "GuildNewsChannel",
    "GuildStoreChannel",
    "GuildVoiceChannel",
    "GuildChannelBuilder",
]

import datetime
import typing

import attr

from hikari import bases
from hikari import permissions
from hikari import users
from hikari.internal import marshaller
from hikari.internal import more_collections
from hikari.internal import more_enums

if typing.TYPE_CHECKING:
    from hikari.internal import more_typing


@more_enums.must_be_unique
class ChannelType(int, more_enums.Enum):
    """The known channel types that are exposed to us by the API."""

    GUILD_TEXT = 0
    """A text channel in a guild."""

    DM = 1
    """A direct channel between two users."""

    GUILD_VOICE = 2
    """A voice channel in a guild."""

    GROUP_DM = 3
    """A direct channel between multiple users."""

    GUILD_CATEGORY = 4
    """An category used for organizing channels in a guild."""

    GUILD_NEWS = 5
    """A channel that can be followed and can crosspost."""

    GUILD_STORE = 6
    """A channel that show's a game's store page."""


@more_enums.must_be_unique
class PermissionOverwriteType(str, more_enums.Enum):
    """The type of entity a Permission Overwrite targets."""

    ROLE = "role"
    """A permission overwrite that targets all the members with a specific role."""

    MEMBER = "member"
    """A permission overwrite that targets a specific guild member."""

    def __str__(self) -> str:
        return self.value


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class PermissionOverwrite(bases.UniqueEntity, marshaller.Deserializable, marshaller.Serializable):
    """Represents permission overwrites for a channel or role in a channel."""

    type: PermissionOverwriteType = marshaller.attrib(deserializer=PermissionOverwriteType, serializer=str)
    """The type of entity this overwrite targets."""

    allow: permissions.Permission = marshaller.attrib(
        deserializer=permissions.Permission, serializer=int, default=permissions.Permission(0)
    )
    """The permissions this overwrite allows."""

    deny: permissions.Permission = marshaller.attrib(
        deserializer=permissions.Permission, serializer=int, default=permissions.Permission(0)
    )
    """The permissions this overwrite denies."""

    @property
    def unset(self) -> permissions.Permission:
        """Bitfield of all permissions not explicitly allowed or denied by this overwrite."""
        return typing.cast(permissions.Permission, (self.allow | self.deny))


def register_channel_type(
    type_: ChannelType,
) -> typing.Callable[[typing.Type[PartialChannel]], typing.Type[PartialChannel]]:
    """Generate a decorator for channel classes defined in this library.

    This allows them to associate themselves with a given channel type.

    Parameters
    ----------
    type_ : ChannelType
        The channel type to associate with.

    Returns
    -------
    decorator(T) -> T
        The decorator to decorate the class with.
    """

    def decorator(cls):
        mapping = getattr(register_channel_type, "types", {})
        mapping[type_] = cls
        setattr(register_channel_type, "types", mapping)
        return cls

    return decorator


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class PartialChannel(bases.UniqueEntity, marshaller.Deserializable):
    """Represents a channel where we've only received it's basic information.

    This is commonly received in REST responses.
    """

    name: typing.Optional[str] = marshaller.attrib(
        deserializer=str, repr=True, default=None, if_undefined=None, if_none=None
    )
    """The channel's name. This will be missing for DM channels."""

    type: ChannelType = marshaller.attrib(deserializer=ChannelType, repr=True)
    """The channel's type."""


def _deserialize_recipients(payload: more_typing.JSONArray, **kwargs: typing.Any) -> typing.Sequence[users.User]:
    return {bases.Snowflake(user["id"]): users.User.deserialize(user, **kwargs) for user in payload}


@register_channel_type(ChannelType.DM)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class DMChannel(PartialChannel):
    """Represents a DM channel."""

    last_message_id: typing.Optional[bases.Snowflake] = marshaller.attrib(deserializer=bases.Snowflake, if_none=None)
    """The ID of the last message sent in this channel.

    !!! note
        This might point to an invalid or deleted message.
    """

    recipients: typing.Mapping[bases.Snowflake, users.User] = marshaller.attrib(
        deserializer=_deserialize_recipients, inherit_kwargs=True,
    )
    """The recipients of the DM."""


@register_channel_type(ChannelType.GROUP_DM)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GroupDMChannel(DMChannel):
    """Represents a DM group channel."""

    owner_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake, repr=True)
    """The ID of the owner of the group."""

    icon_hash: typing.Optional[str] = marshaller.attrib(raw_name="icon", deserializer=str, if_none=None)
    """The hash of the icon of the group."""

    application_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake, if_undefined=None, default=None
    )
    """The ID of the application that created the group DM, if it's a bot based group DM."""


def _deserialize_overwrites(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Mapping[bases.Snowflake, PermissionOverwrite]:
    return {
        bases.Snowflake(overwrite["id"]): PermissionOverwrite.deserialize(overwrite, **kwargs) for overwrite in payload
    }


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildChannel(PartialChannel):
    """The base for anything that is a guild channel."""

    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake, if_undefined=None, default=None, repr=True
    )
    """The ID of the guild the channel belongs to.

    This will be `None` when received over the gateway in certain events (e.g.
    Guild Create).
    """

    position: int = marshaller.attrib(deserializer=int)
    """The sorting position of the channel."""

    permission_overwrites: PermissionOverwrite = marshaller.attrib(
        deserializer=_deserialize_overwrites, inherit_kwargs=True
    )
    """The permission overwrites for the channel."""

    is_nsfw: typing.Optional[bool] = marshaller.attrib(
        raw_name="nsfw", deserializer=bool, if_undefined=None, default=None
    )
    """Whether the channel is marked as NSFW.

    This will be `None` when received over the gateway in certain events (e.g
    Guild Create).
    """

    parent_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake, if_none=None, if_undefined=None, repr=True
    )
    """The ID of the parent category the channel belongs to."""


@register_channel_type(ChannelType.GUILD_CATEGORY)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildCategory(GuildChannel):
    """Represents a guild category."""


def _deserialize_rate_limit_per_user(payload: int) -> datetime.timedelta:
    return datetime.timedelta(seconds=payload)


@register_channel_type(ChannelType.GUILD_TEXT)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildTextChannel(GuildChannel):
    """Represents a guild text channel."""

    topic: typing.Optional[str] = marshaller.attrib(deserializer=str, if_none=None)
    """The topic of the channel."""

    last_message_id: typing.Optional[bases.Snowflake] = marshaller.attrib(deserializer=bases.Snowflake, if_none=None)
    """The ID of the last message sent in this channel.

    !!! note
        This might point to an invalid or deleted message.
    """

    rate_limit_per_user: datetime.timedelta = marshaller.attrib(deserializer=_deserialize_rate_limit_per_user)
    """The delay (in seconds) between a user can send a message to this channel.

    !!! note
        Bots, as well as users with `MANAGE_MESSAGES` or `MANAGE_CHANNEL`,
        are not affected by this.
    """


@register_channel_type(ChannelType.GUILD_NEWS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildNewsChannel(GuildChannel):
    """Represents an news channel."""

    topic: str = marshaller.attrib(deserializer=str, if_none=None)
    """The topic of the channel."""

    last_message_id: typing.Optional[bases.Snowflake] = marshaller.attrib(deserializer=bases.Snowflake, if_none=None)
    """The ID of the last message sent in this channel.

    !!! note
        This might point to an invalid or deleted message.
    """


@register_channel_type(ChannelType.GUILD_STORE)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildStoreChannel(GuildChannel):
    """Represents a store channel."""


@register_channel_type(ChannelType.GUILD_VOICE)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildVoiceChannel(GuildChannel):
    """Represents an voice channel."""

    bitrate: int = marshaller.attrib(deserializer=int, repr=True)
    """The bitrate for the voice channel (in bits)."""

    user_limit: int = marshaller.attrib(deserializer=int, repr=True)
    """The user limit for the voice channel."""


class GuildChannelBuilder(marshaller.Serializable):
    """Used to create channel objects to send in guild create requests.

    Parameters
    ----------
    channel_name : str
        The name to set for the channel.
    channel_type : ChannelType
        The type of channel this should build.

    Examples
    --------
        channel_obj = (
            channels.GuildChannelBuilder("Catgirl-appreciation", channels.ChannelType.GUILD_TEXT)
            .is_nsfw(True)
            .with_topic("Here we men of culture appreciate the way of the neko.")
            .with_rate_limit_per_user(datetime.timedelta(seconds=5))
            .with_permission_overwrites([overwrite_obj])
            .with_id(1)
        )
    """

    __slots__ = ("_payload",)

    def __init__(self, channel_name: str, channel_type: ChannelType) -> None:
        self._payload: typing.Dict[str, typing.Any] = {
            "type": channel_type,
            "name": channel_name,
        }

    def serialize(self: GuildChannelBuilder) -> typing.Mapping[str, typing.Any]:
        """Serialize this instance into a payload to send to Discord."""
        return self._payload

    def is_nsfw(self) -> GuildChannelBuilder:
        """Mark this channel as NSFW."""
        self._payload["nsfw"] = True
        return self

    def with_permission_overwrites(self, overwrites: typing.Sequence[PermissionOverwrite]) -> GuildChannelBuilder:
        """Set the permission overwrites for this channel.

        Parameters
        ----------
        overwrites : typing.Sequence[PermissionOverwrite]
            A sequence of overwrite objects to add, where the first overwrite
            object

        !!! note
            Calling this multiple times will overwrite any previously added
            overwrites.
        """
        self._payload["permission_overwrites"] = [o.serialize() for o in overwrites]
        return self

    def with_topic(self, topic: str) -> GuildChannelBuilder:
        """Set the topic for this channel.

        Parameters
        ----------
        topic : str
            The string topic to set.
        """
        self._payload["topic"] = topic
        return self

    def with_bitrate(self, bitrate: int) -> GuildChannelBuilder:
        """Set the bitrate for this channel.

        Parameters
        ----------
        bitrate : int
            The bitrate to set in bits.
        """
        self._payload["bitrate"] = int(bitrate)
        return self

    def with_user_limit(self, user_limit: int) -> GuildChannelBuilder:
        """Set the limit for how many users can be in this channel at once.

        Parameters
        ----------
        user_limit : int
            The user limit to set.
        """
        self._payload["user_limit"] = int(user_limit)
        return self

    def with_rate_limit_per_user(
        self, rate_limit_per_user: typing.Union[datetime.timedelta, int]
    ) -> GuildChannelBuilder:
        """Set the rate limit for users sending messages in this channel.

        Parameters
        ----------
        rate_limit_per_user : typing.Union[datetime.timedelta, int]
            The amount of seconds users will have to wait before sending another
            message in the channel to set.
        """
        self._payload["rate_limit_per_user"] = int(
            rate_limit_per_user.total_seconds()
            if isinstance(rate_limit_per_user, datetime.timedelta)
            else rate_limit_per_user
        )
        return self

    def with_parent_category(self, category: typing.Union[bases.Snowflake, int]) -> GuildChannelBuilder:
        """Set the parent category for this channel.

        Parameters
        ----------
        category : typing.Union[hikari.bases.Snowflake, int]
            The placeholder ID of the category channel that should be this
            channel's parent.
        """
        self._payload["parent_id"] = str(int(category))
        return self

    def with_id(self, channel_id: typing.Union[bases.Snowflake, int]) -> GuildChannelBuilder:
        """Set the placeholder ID for this channel.

        Parameters
        ----------
        channel_id : typing.Union[hikari.bases.Snowflake, int]
            The placeholder ID to use.

        !!! note
            This ID is purely a place holder used for setting parent category
            channels and will have no effect on the created channel's ID.
        """
        self._payload["id"] = str(int(channel_id))
        return self


def deserialize_channel(payload: more_typing.JSONObject, **kwargs: typing.Any) -> typing.Union[GuildChannel, DMChannel]:
    """Deserialize a channel object into the corresponding class.

    !!! warning
        This can only be used to deserialize full channel objects. To
        deserialize a partial object, use `PartialChannel.deserialize`.
    """
    type_id = payload["type"]
    types = getattr(register_channel_type, "types", more_collections.EMPTY_DICT)
    channel_type = types[type_id]
    return channel_type.deserialize(payload, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Datastructure bases."""

from __future__ import annotations

__all__ = ["HikariEntity", "Snowflake", "UniqueEntity"]

import abc
import typing

import attr

from hikari.internal import conversions
from hikari.internal import marshaller

if typing.TYPE_CHECKING:
    import datetime

    from hikari.clients import components


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True, init=False)
class HikariEntity(abc.ABC):
    """The base for any entity used in this API."""

    _components: typing.Optional[components.Components] = attr.attrib(default=None, repr=False, eq=False, hash=False)
    """The client components that models may use for procedures."""


class Snowflake(int):
    """A concrete representation of a unique identifier for an object on Discord.

    This object can be treated as a regular `int` for most purposes.
    """

    __slots__ = ()

    ___MIN___: Snowflake
    ___MAX___: Snowflake

    @staticmethod
    def __new__(cls, value: typing.Union[int, str]) -> Snowflake:
        return super(Snowflake, cls).__new__(cls, value)

    @property
    def created_at(self) -> datetime.datetime:
        """When the object was created."""
        epoch = self >> 22
        return conversions.discord_epoch_to_datetime(epoch)

    @property
    def internal_worker_id(self) -> int:
        """ID of the worker that created this snowflake on Discord's systems."""
        return (self & 0x3E0_000) >> 17

    @property
    def internal_process_id(self) -> int:
        """ID of the process that created this snowflake on Discord's systems."""
        return (self & 0x1F_000) >> 12

    @property
    def increment(self) -> int:
        """Increment of Discord's system when this object was made."""
        return self & 0xFFF

    def serialize(self) -> str:
        """Generate a JSON-friendly representation of this object."""
        return str(self)

    @classmethod
    def from_datetime(cls, date: datetime.datetime) -> Snowflake:
        """Get a snowflake object from a datetime object."""
        return cls.from_timestamp(date.timestamp())

    @classmethod
    def from_timestamp(cls, timestamp: float) -> Snowflake:
        """Get a snowflake object from a UNIX timestamp."""
        return cls(int(timestamp - conversions.DISCORD_EPOCH) * 1000 << 22)

    @classmethod
    def min(cls) -> Snowflake:
        """Minimum value for a snowflake."""
        if not hasattr(cls, "___MIN___"):
            cls.___MIN___ = Snowflake(0)
        return cls.___MIN___

    @classmethod
    def max(cls) -> Snowflake:
        """Maximum value for a snowflake."""
        if not hasattr(cls, "___MAX___"):
            cls.___MAX___ = Snowflake((1 << 63) - 1)
        return cls.___MAX___


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class UniqueEntity(HikariEntity, typing.SupportsInt, abc.ABC):
    """A base for an entity that has an integer ID of some sort.

    Casting an object of this type to an `int` will produce the
    integer ID of the object.
    """

    id: Snowflake = marshaller.attrib(hash=True, eq=True, repr=True, deserializer=Snowflake, serializer=str)
    """The ID of this entity."""

    def __int__(self) -> int:
        return int(self.id)


T = typing.TypeVar("T", bound=UniqueEntity)
Hashable = typing.Union[Snowflake, int, T]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Various reusable type-hints for this library."""
# pylint:disable=unused-variable

from __future__ import annotations

__all__ = [
    "JSONType",
    "NullableJSONArray",
    "JSONObject",
    "NullableJSONObject",
    "JSONArray",
    "NullableJSONType",
    "Headers",
    "Coroutine",
    "Future",
    "Task",
]

# Hide any imports; this encourages any uses of this to use the typing module
# for regular stuff rather than relying on it being in here as well.
# pylint: disable=ungrouped-imports
from typing import Any as _Any
from typing import AnyStr as _AnyStr
from typing import Coroutine as _Coroutine
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Protocol as _Protocol
from typing import runtime_checkable as _runtime_checkable
from typing import Sequence as _Sequence
from typing import TYPE_CHECKING as _TYPE_CHECKING
from typing import TypeVar as _TypeVar
from typing import Union as _Union

if _TYPE_CHECKING:
    from types import FrameType as _FrameType
    from typing import Callable as _Callable
    from typing import IO as _IO
    import asyncio
    import contextvars
# pylint: enable=ungrouped-imports

T_contra = _TypeVar("T_contra", contravariant=True)
# noinspection PyShadowingBuiltins
T_co = _TypeVar("T_co", covariant=True)

##########################
# HTTP TYPE HINT HELPERS #
##########################

JSONType = _Union[
    _Mapping[str, "NullableJSONType"], _Sequence["NullableJSONType"], _AnyStr, int, float, bool,
]
"""Any JSON type."""

NullableJSONType = _Optional[JSONType]
"""Any JSON type, including `null`."""

JSONObject = _Mapping[str, NullableJSONType]
"""A mapping produced from a JSON object."""

NullableJSONObject = _Optional[JSONObject]
"""A mapping produced from a JSON object that may or may not be present."""

JSONArray = _Sequence[NullableJSONType]
"""A sequence produced from a JSON array."""

NullableJSONArray = _Optional[JSONArray]
"""A sequence produced from a JSON array that may or may not be present."""

Headers = _Mapping[str, _Union[_Sequence[str], str]]
"""HTTP headers."""

#############################
# ASYNCIO TYPE HINT HELPERS #
#############################

Coroutine = _Coroutine[_Any, _Any, T_co]
"""A coroutine object.

This is awaitable but MUST be awaited somewhere to be completed correctly.
"""


@_runtime_checkable
class Future(_Protocol[T_contra]):
    """Typed protocol representation of an `asyncio.Future`.

    You should consult the documentation for `asyncio.Future` for usage.
    """

    def result(self) -> T_contra:
        """See `asyncio.Future.result`."""

    def set_result(self, result: T_contra, /) -> None:
        """See `asyncio.Future.set_result`."""

    def set_exception(self, exception: Exception, /) -> None:
        """See `asyncio.Future.set_exception`."""

    def done(self) -> bool:
        """See `asyncio.Future.done`."""

    def cancelled(self) -> bool:
        """See `asyncio.Future.cancelled`."""

    def add_done_callback(
        self, callback: _Callable[[Future[T_contra]], None], /, *, context: _Optional[contextvars.Context],
    ) -> None:
        """See `asyncio.Future.add_done_callback`."""

    def remove_done_callback(self, callback: _Callable[[Future[T_contra]], None], /) -> None:
        """See `asyncio.Future.remove_done_callback`."""

    def cancel(self) -> bool:
        """See `asyncio.Future.cancel`."""

    def exception(self) -> _Optional[Exception]:
        """See `asyncio.Future.exception`."""

    def get_loop(self) -> asyncio.AbstractEventLoop:
        """See `asyncio.Future.get_loop`."""

    def __await__(self) -> Coroutine[T_contra]:
        ...


@_runtime_checkable
class Task(_Protocol[T_contra]):
    """Typed protocol representation of an `asyncio.Task`.

    You should consult the documentation for `asyncio.Task` for usage.
    """

    def result(self) -> T_contra:
        """See`asyncio.Future.result`."""

    def set_result(self, result: T_contra, /) -> None:
        """See `asyncio.Future.set_result`."""

    def set_exception(self, exception: Exception, /) -> None:
        """See `asyncio.Future.set_exception`."""

    def done(self) -> bool:
        """See `asyncio.Future.done`."""

    def cancelled(self) -> bool:
        """See `asyncio.Future.cancelled`."""

    def add_done_callback(
        self, callback: _Callable[[Future[T_contra]], None], /, *, context: _Optional[contextvars.Context],
    ) -> None:
        """See `asyncio.Future.add_done_callback`."""

    def remove_done_callback(self, callback: _Callable[[Future[T_contra]], None], /) -> None:
        """See `asyncio.Future.remove_done_callback`."""

    def cancel(self) -> bool:
        """See `asyncio.Future.cancel`."""

    def exception(self) -> _Optional[Exception]:
        """See `asyncio.Future.exception`."""

    def get_loop(self) -> asyncio.AbstractEventLoop:
        """See `asyncio.Future.get_loop`."""

    def get_stack(self, *, limit: _Optional[int] = None) -> _Sequence[_FrameType]:
        """See `asyncio.Task.get_stack`."""

    def print_stack(self, *, limit: _Optional[int] = None, file: _Optional[_IO] = None) -> None:
        """See `asyncio.Task.print_stack`."""

    def get_name(self) -> str:
        """See `asyncio.Task.get_name`."""

    def set_name(self, value: str, /) -> None:
        """See `asyncio.Task.set_name`."""

    def __await__(self) -> Coroutine[T_contra]:
        ...


# pylint:enable=unused-variable



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities related to discord's Oauth2 flow."""

from __future__ import annotations

__all__ = [
    "Application",
    "ApplicationOwner",
    "ConnectionVisibility",
    "OAuth2Scope",
    "OwnConnection",
    "OwnGuild",
    "Team",
    "TeamMember",
    "TeamMembershipState",
]

import typing

import attr

from hikari import bases
from hikari import guilds
from hikari import permissions
from hikari import users
from hikari.internal import marshaller
from hikari.internal import more_enums
from hikari.internal import urls

if typing.TYPE_CHECKING:
    from hikari.internal import more_typing


@more_enums.must_be_unique
class OAuth2Scope(str, more_enums.Enum):
    """OAuth2 Scopes that Discord allows.

    These are categories of permissions for applications using the OAuth2 API
    directly. Most users will only ever need the `BOT` scope when developing
    bots.
    """

    ACTIVITIES_READ = "activities.read"
    """Enable the app to fetch a user's "Now Playing/Recently Played" list.

    !!! note:
        You must be whitelisted to use this scope.
    """

    ACTIVITIES_WRITE = "activities.write"
    """Enable the app to update a user's activity.

    !!! note:
        You must be whitelisted to use this scope.

    !!! note:
        This is not required to use the GameSDK activity manager.
    """

    APPLICATIONS_BUILDS_READ = "applications.builds.read"
    """Enable the app to read build data for a user's applications.

    !!! note:
        You must be whitelisted to use this scope.
    """

    APPLICATIONS_BUILDS_UPLOAD = "applications.builds.upload"
    """Enable the app to upload/update builds for a user's applications.

    !!! note:
        You must be whitelisted to use this scope.
    """

    APPLICATIONS_ENTITLEMENTS = "applications.entitlements"
    """Enable the app to read entitlements for a user's applications."""

    APPLICATIONS_STORE_UPDATE = "applications.store.update"
    """Enable the app to read and update store data for the user's applications.

    This includes store listings, achievements, SKU's, etc.

    !!! note:
        The store API is deprecated and may be removed in the future.
    """

    BOT = "bot"
    """Used to add OAuth2 bots to a guild.

    !!! note:
        This requires you to have set up a bot account for your application.
    """

    CONNECTIONS = "connections"
    """Enable the app to view third-party linked accounts such as Twitch."""

    EMAIL = "email"
    """Enable the app to view the user's email and application info."""

    GROUP_DM_JOIN = "gdm.join"
    """Enable the application to join users into a group DM."""

    GUILDS = "guilds"
    """Enable the app to view the guilds the user is in."""

    GUILDS_JOIN = "guilds.join"
    """Enable the app to add the user to a specific guild.

    !!! note:
        This requires you to have set up a bot account for your application.
    """

    IDENTIFY = "identify"
    """Enable the app to view info about itself.

    !!! note:
        This does not include email address info. Use the `EMAIL` scope instead
        to retrieve this information.
    """

    RELATIONSHIPS_READ = "relationships.read"
    """Enable the app to view a user's friend list.

    !!! note:
        You must be whitelisted to use this scope.
    """

    RPC = "rpc"
    """Enable the RPC app to control the local user's Discord client.

    !!! note:
        You must be whitelisted to use this scope.
    """

    RPC_API = "rpc.api"
    """Enable the RPC app to access the RPC API as the local user.

    !!! note:
        You must be whitelisted to use this scope.
    """

    RPC_MESSAGES_READ = "messages.read"
    """Enable the RPC app to read messages from all channels the user is in."""

    RPC_NOTIFICATIONS_READ = "rpc.notifications.read"
    """Enable the RPC app to read  from all channels the user is in.

    !!! note:
        You must be whitelisted to use this scope.
    """

    WEBHOOK_INCOMING = "webhook.incoming"
    """Used to generate a webhook that is returned in the OAuth2 token response.

    This is used during authorization code grants.
    """


@more_enums.must_be_unique
class ConnectionVisibility(int, more_enums.Enum):
    """Describes who can see a connection with a third party account."""

    NONE = 0
    """Only you can see the connection."""

    EVERYONE = 1
    """Everyone can see the connection."""


def _deserialize_integrations(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Sequence[guilds.GuildIntegration]:
    return [guilds.PartialGuildIntegration.deserialize(integration, **kwargs) for integration in payload]


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class OwnConnection(bases.HikariEntity, marshaller.Deserializable):
    """Represents a user's connection with a third party account.

    Returned by the `GET Current User Connections` endpoint.
    """

    id: str = marshaller.attrib(deserializer=str)
    """The string ID of the third party connected account.

    !!! warning
        Seeing as this is a third party ID, it will not be a snowflake.
    """

    name: str = marshaller.attrib(deserializer=str, repr=True)
    """The username of the connected account."""

    type: str = marshaller.attrib(deserializer=str, repr=True)
    """The type of service this connection is for."""

    is_revoked: bool = marshaller.attrib(raw_name="revoked", deserializer=bool, if_undefined=False, default=False)
    """Whether the connection has been revoked."""

    integrations: typing.Sequence[guilds.PartialGuildIntegration] = marshaller.attrib(
        deserializer=_deserialize_integrations, if_undefined=list, factory=list, inherit_kwargs=True,
    )
    """A sequence of the partial guild integration objects this connection has."""

    is_verified: bool = marshaller.attrib(raw_name="verified", deserializer=bool)
    """Whether the connection has been verified."""

    is_friend_syncing: bool = marshaller.attrib(raw_name="friend_sync", deserializer=bool)
    """Whether friends should be added based on this connection."""

    is_showing_activity: bool = marshaller.attrib(raw_name="show_activity", deserializer=bool)
    """Whether this connection's activities are shown in the user's presence."""

    visibility: ConnectionVisibility = marshaller.attrib(deserializer=ConnectionVisibility, repr=True)
    """The visibility of the connection."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class OwnGuild(guilds.PartialGuild):
    """Represents a user bound partial guild object."""

    is_owner: bool = marshaller.attrib(raw_name="owner", deserializer=bool, repr=True)
    """Whether the current user owns this guild."""

    my_permissions: permissions.Permission = marshaller.attrib(
        raw_name="permissions", deserializer=permissions.Permission
    )
    """The guild level permissions that apply to the current user or bot."""


@more_enums.must_be_unique
class TeamMembershipState(int, more_enums.Enum):
    """Represents the state of a user's team membership."""

    INVITED = 1
    """Denotes the user has been invited to the team but has yet to accept."""

    ACCEPTED = 2
    """Denotes the user has accepted the invite and is now a member."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class TeamMember(bases.HikariEntity, marshaller.Deserializable):
    """Represents a member of a Team."""

    membership_state: TeamMembershipState = marshaller.attrib(deserializer=TeamMembershipState)
    """The state of this user's membership."""

    permissions: typing.Set[str] = marshaller.attrib(deserializer=set)
    """This member's permissions within a team.

    Will always be `["*"]` until Discord starts using this.
    """

    team_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake, repr=True)
    """The ID of the team this member belongs to."""

    user: users.User = marshaller.attrib(deserializer=users.User.deserialize, inherit_kwargs=True, repr=True)
    """The user object of this team member."""


def _deserialize_members(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Mapping[bases.Snowflake, TeamMember]:
    return {bases.Snowflake(member["user"]["id"]): TeamMember.deserialize(member, **kwargs) for member in payload}


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class Team(bases.UniqueEntity, marshaller.Deserializable):
    """Represents a development team, along with all its members."""

    icon_hash: typing.Optional[str] = marshaller.attrib(raw_name="icon", deserializer=str)
    """The hash of this team's icon, if set."""

    members: typing.Mapping[bases.Snowflake, TeamMember] = marshaller.attrib(
        deserializer=_deserialize_members, inherit_kwargs=True
    )
    """The member's that belong to this team."""

    owner_user_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake, repr=True)
    """The ID of this team's owner."""

    @property
    def icon_url(self) -> typing.Optional[str]:
        """URL of this team's icon, if set."""
        return self.format_icon_url()

    def format_icon_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the icon URL for this team if set.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png`.
            Supports `png`, `jpeg`, `jpg` and `webp`.
        size : int
            The size to set for the URL, defaults to `4096`. Can be any power
            of two between `16` and `4096` inclusive.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.icon_hash:
            return urls.generate_cdn_url("team-icons", str(self.id), self.icon_hash, fmt=fmt, size=size)
        return None


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ApplicationOwner(users.User):
    """Represents the user who owns an application, may be a team user."""

    flags: int = marshaller.attrib(deserializer=users.UserFlag, repr=True)
    """This user's flags."""

    @property
    def is_team_user(self) -> bool:
        """If this user is a Team user (the owner of an application that's owned by a team)."""
        return bool((self.flags >> 10) & 1)


def _deserialize_verify_key(payload: str) -> bytes:
    return bytes(payload, "utf-8")


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class Application(bases.UniqueEntity, marshaller.Deserializable):
    """Represents the information of an Oauth2 Application."""

    name: str = marshaller.attrib(deserializer=str, repr=True)
    """The name of this application."""

    description: str = marshaller.attrib(deserializer=str)
    """The description of this application, will be an empty string if unset."""

    is_bot_public: typing.Optional[bool] = marshaller.attrib(
        raw_name="bot_public", deserializer=bool, if_undefined=None, default=None, repr=True
    )
    """Whether the bot associated with this application is public.

    Will be `None` if this application doesn't have an associated bot.
    """

    is_bot_code_grant_required: typing.Optional[bool] = marshaller.attrib(
        raw_name="bot_require_code_grant", deserializer=bool, if_undefined=None, default=None
    )
    """Whether this application's bot is requiring code grant for invites.

    Will be `None` if this application doesn't have a bot.
    """

    owner: typing.Optional[ApplicationOwner] = marshaller.attrib(
        deserializer=ApplicationOwner.deserialize, if_undefined=None, default=None, inherit_kwargs=True, repr=True
    )
    """The object of this application's owner.

    This should always be `None` in application objects retrieved outside
    Discord's oauth2 flow.
    """

    rpc_origins: typing.Optional[typing.Set[str]] = marshaller.attrib(deserializer=set, if_undefined=None, default=None)
    """A collection of this application's rpc origin URLs, if rpc is enabled."""

    summary: str = marshaller.attrib(deserializer=str)
    """This summary for this application's primary SKU if it's sold on Discord.

    Will be an empty string if unset.
    """

    verify_key: typing.Optional[bytes] = marshaller.attrib(
        deserializer=_deserialize_verify_key, if_undefined=None, default=None
    )
    """The base64 encoded key used for the GameSDK's `GetTicket`."""

    icon_hash: typing.Optional[str] = marshaller.attrib(
        raw_name="icon", deserializer=str, if_undefined=None, default=None
    )
    """The hash of this application's icon, if set."""

    team: typing.Optional[Team] = marshaller.attrib(
        deserializer=Team.deserialize, if_undefined=None, if_none=None, default=None, inherit_kwargs=True
    )
    """This application's team if it belongs to one."""

    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake, if_undefined=None, default=None
    )
    """The ID of the guild this application is linked to if sold on Discord."""

    primary_sku_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake, if_undefined=None, default=None
    )
    """The ID of the primary "Game SKU" of a game that's sold on Discord."""

    slug: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None)
    """The URL slug that links to this application's store page.

    Only applicable to applications sold on Discord.
    """

    cover_image_hash: typing.Optional[str] = marshaller.attrib(
        raw_name="cover_image", deserializer=str, if_undefined=None, default=None
    )
    """The hash of this application's cover image on it's store, if set."""

    @property
    def icon_url(self) -> typing.Optional[str]:
        """URL for this team's icon, if set."""
        return self.format_icon_url()

    def format_icon_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the icon URL for this application if set.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png`.
            Supports `png`, `jpeg`, `jpg` and `webp`.
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.icon_hash:
            return urls.generate_cdn_url("app-icons", str(self.id), self.icon_hash, fmt=fmt, size=size)
        return None

    @property
    def cover_image_url(self) -> typing.Optional[str]:
        """URL for this icon's store cover image, if set."""
        return self.format_cover_image_url()

    def format_cover_image_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the URL for this application's store page's cover image is set and applicable.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png`.
            Supports `png`, `jpeg`, `jpg` and `webp`.
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.cover_image_hash:
            return urls.generate_cdn_url("app-assets", str(self.id), self.cover_image_hash, fmt=fmt, size=size)
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Mixin utilities for defining enums."""

from __future__ import annotations

__all__ = ["Enum", "IntFlag", "must_be_unique", "generated_value"]

import enum
import typing


class Enum(enum.Enum):
    """A non-flag enum type.

    This gives a more meaningful `__str__` implementation than what is defined
    in the `enum` module by default.
    """

    __slots__ = ()

    name: str
    """The name of the enum member."""

    def __str__(self) -> str:
        return self.name


class IntFlag(enum.IntFlag):
    """Base for an integer flag enum type.

    This gives a more meaningful `__str__` implementation than what is defined
    in the `enum` module by default.
    """

    __slots__ = ()

    name: str
    """The name of the enum member."""

    def __str__(self) -> str:
        return ", ".join(flag.name for flag in typing.cast(typing.Iterable, type(self)) if flag & self)


must_be_unique = enum.unique
generated_value = enum.auto



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""The logic for handling requests to `@me` endpoints."""

from __future__ import annotations

__all__ = ["RESTCurrentUserComponent"]

import abc
import datetime
import functools
import typing

from hikari import applications
from hikari import bases
from hikari import channels as _channels
from hikari import users
from hikari.clients.rest import base
from hikari.internal import helpers

if typing.TYPE_CHECKING:
    from hikari import guilds
    from hikari import files


class RESTCurrentUserComponent(base.BaseRESTComponent, abc.ABC):  # pylint: disable=abstract-method
    """The REST client component for handling requests to `@me` endpoints."""

    async def fetch_me(self) -> users.MyUser:
        """Get the current user that of the token given to the client.

        Returns
        -------
        hikari.users.MyUser
            The current user object.
        """
        payload = await self._session.get_current_user()
        return users.MyUser.deserialize(payload, components=self._components)

    async def update_me(self, *, username: str = ..., avatar: typing.Optional[files.BaseStream] = ...) -> users.MyUser:
        """Edit the current user.

        Parameters
        ----------
        username : str
            If specified, the new username string.
        avatar : hikari.files.BaseStream, optional
            If specified, the new avatar image data.
            If it is None, the avatar is removed.

        !!! warning
            Verified bots will not be able to change their username on this
            endpoint, and should contact Discord support instead to change
            this value.

        Returns
        -------
        hikari.users.MyUser
            The updated user object.

        Raises
        ------
        hikari.errors.BadRequest
            If you pass username longer than the limit (`2-32`) or an invalid image.
        """
        payload = await self._session.modify_current_user(
            username=username, avatar=await avatar.read() if avatar is not ... else ...,
        )
        return users.MyUser.deserialize(payload, components=self._components)

    async def fetch_my_connections(self) -> typing.Sequence[applications.OwnConnection]:
        """
        Get the current user's connections.

        !!! note
            This endpoint can be used with both `Bearer` and `Bot` tokens but
            will usually return an empty list for bots (with there being some
            exceptions to this, like user accounts that have been converted to
            bots).

        Returns
        -------
        typing.Sequence[hikari.applications.OwnConnection]
            A list of connection objects.
        """
        payload = await self._session.get_current_user_connections()
        return [
            applications.OwnConnection.deserialize(connection, components=self._components) for connection in payload
        ]

    def fetch_my_guilds_after(
        self,
        *,
        after: typing.Union[datetime.datetime, bases.Hashable[guilds.Guild]] = 0,
        limit: typing.Optional[int] = None,
    ) -> typing.AsyncIterator[applications.OwnGuild]:
        """Get an async iterator of the guilds the current user is in.

        This returns the guilds created after a given guild object/ID or from
        the oldest guild.

        Parameters
        ----------
        after : typing.Union[datetime.datetime, hikari.guilds.Guild, hikari.bases.Snowflake, int]
            The object or ID of a guild to get guilds that were created after
            it if specified, else this will start at the oldest guild.
        limit : int
            If specified, the maximum amount of guilds that this paginator
            should return.

        Examples
        --------
            async for user in client.fetch_my_guilds_after(after=9876543, limit=1231):
                await client.leave_guild(guild)

        Returns
        -------
        typing.AsyncIterator[hikari.applications.OwnGuild]
            An async iterator of partial guild objects.

        Raises
        ------
        hikari.errors.NotFound
            If the guild is not found.
        hikari.errors.BadRequest
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        """
        if isinstance(after, datetime.datetime):
            after = str(bases.Snowflake.from_datetime(after))
        else:
            after = str(after.id if isinstance(after, bases.UniqueEntity) else int(after))
        deserializer = functools.partial(applications.OwnGuild.deserialize, components=self._components)
        return helpers.pagination_handler(
            deserializer=deserializer,
            direction="after",
            request=self._session.get_current_user_guilds,
            reversing=False,
            start=after,
            maximum_limit=100,
            limit=limit,
        )

    def fetch_my_guilds_before(
        self,
        *,
        before: typing.Union[datetime.datetime, bases.Hashable[guilds.Guild]] = bases.Snowflake.max(),
        limit: typing.Optional[int] = None,
    ) -> typing.AsyncIterator[applications.OwnGuild]:
        """Get an async iterator of the guilds the current user is in.

        This returns the guilds that were created before a given user object/ID
        or from the newest guild.

        Parameters
        ----------
        before : typing.Union[datetime.datetime, hikari.guilds.Guild, hikari.bases.Snowflake, int]
            The object or ID of a guild to get guilds that were created
            before it if specified, else this will start at the newest guild.
        limit : int
            If specified, the maximum amount of guilds that this paginator
            should return.

        Returns
        -------
        typing.AsyncIterator[hikari.applications.OwnGuild]
            An async iterator of partial guild objects.

        Raises
        ------
        hikari.errors.NotFound
            If the guild is not found.
        hikari.errors.BadRequest
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        """
        if isinstance(before, datetime.datetime):
            before = str(bases.Snowflake.from_datetime(before))
        else:
            before = str(before.id if isinstance(before, bases.UniqueEntity) else int(before))
        deserializer = functools.partial(applications.OwnGuild.deserialize, components=self._components)
        return helpers.pagination_handler(
            deserializer=deserializer,
            direction="before",
            request=self._session.get_current_user_guilds,
            reversing=False,
            start=before,
            maximum_limit=100,
            limit=limit,
        )

    async def leave_guild(self, guild: bases.Hashable[guilds.Guild]) -> None:
        """Make the current user leave a given guild.

        Parameters
        ----------
        guild : typing.Union[hikari.guilds.Guild, hikari.bases.Snowflake, int]
            The object or ID of the guild to leave.

        Raises
        ------
        hikari.errors.NotFound
            If the guild is not found.
        hikari.errors.BadRequest
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        """
        await self._session.leave_guild(guild_id=str(guild.id if isinstance(guild, bases.UniqueEntity) else int(guild)))

    async def create_dm_channel(self, recipient: bases.Hashable[users.User]) -> _channels.DMChannel:
        """Create a new DM channel with a given user.

        Parameters
        ----------
        recipient : typing.Union[hikari.users.User, hikari.bases.Snowflake, int]
            The object or ID of the user to create the new DM channel with.

        Returns
        -------
        hikari.channels.DMChannel
            The newly created DM channel object.

        Raises
        ------
        hikari.errors.NotFound
            If the recipient is not found.
        hikari.errors.BadRequest
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        """
        payload = await self._session.create_dm(
            recipient_id=str(recipient.id if isinstance(recipient, bases.UniqueEntity) else int(recipient))
        )
        return _channels.DMChannel.deserialize(payload, components=self._components)
